/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.portal.security.impl.jacc;

import java.io.FileInputStream;
import java.security.CodeSource;
import java.security.Permission;
import java.security.PermissionCollection;
import java.security.Policy;
import java.security.Principal;
import java.security.ProtectionDomain;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.security.auth.Subject;
import javax.security.jacc.EJBMethodPermission;
import javax.security.jacc.EJBRoleRefPermission;
import javax.security.jacc.PolicyConfiguration;
import javax.security.jacc.PolicyContext;
import javax.security.jacc.PolicyContextException;
import javax.security.jacc.WebResourcePermission;
import javax.security.jacc.WebRoleRefPermission;
import javax.security.jacc.WebUserDataPermission;

import org.jboss.logging.Logger;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentReaderHashMap;

/**
 * RipOff of the AppServer's DelegatingPolicy.
 * <p>
 * This is necessary to avoid redeployment issues with the access control
 * checks. The DelegatingPolicy of the Application Server should be able to
 * handle this, but I was not able to get it to work properly across
 * redeployments. Idealy, the policy and the security service are defined
 * external to the portal sar, but that raised other issues with the
 * instatiation of the DelegatingPolicy and the external permission classes
 * </p>
 * 
 * <p>
 * Added some logic to demontstrate silent authentication to the portal framework objects.
 * Illustrates how one can add custom security checks or replace the existing
 * security calls to the portal db tables with a pure JACC implementation.
 * </p>
 * 
 * @author Scott.Stark@jboss.org
 * @author <a href="mailto:mholzner@novell.com">Martin Holzner</a>
 * @auther James.Williams@jboss.com
 */
public class PortalJaccPolicy extends Policy {
	private static Logger log = Logger.getLogger(PortalJaccPolicy.class);

	private static final String SILENT_AUTHORIZATION_STORE = "C:\\silent-auth.properties";
	
	private static PortalJaccPolicy instance;
	
	/**
	 * The system Policy we delegate non-JACC checks to
	 */
	private Policy delegate;

	/**
	 * Map<String, ContextPolicy> for the JACC context IDs that have been
	 * committed.
	 */
	private ConcurrentReaderHashMap activePolicies = new ConcurrentReaderHashMap();

	/**
	 * Map<String, ContextPolicy> for the JACC policies that are in the open
	 * state and should be excluded from the active permission set.
	 */
	private ConcurrentReaderHashMap openPolicies = new ConcurrentReaderHashMap();

	/**
	 * Flag indicating if our category is at trace level for logging
	 */
	private boolean trace;

	/**
	 * The Policy proxy returned via the PolicyProxy attribute
	 */
	// private PolicyProxy policyProxy = new PolicyProxy(this);
	/**
	 * The types of permissions which should be treated as JACC permission types
	 * in terms of whether this policy should validate the permission.
	 */
	private Class[] externalPermissionTypes = {};

	public synchronized static Policy getInstance() {
		if (instance == null) {
			instance = new PortalJaccPolicy();
		}
		return instance;
	}

	public PortalJaccPolicy() {
		this(null);
	}

	public PortalJaccPolicy(Policy delegate) {
		log.info("Loading up custom PortalJaccPolicy");
		if (delegate == null) {
			delegate = Policy.getPolicy();
		}
		this.delegate = delegate;
		this.trace = log.isTraceEnabled();
		if (instance == null) {
			instance = this;
		}

		/*
		 * When run with a security manager the act of class loading can trigger
		 * security checks which in turn causes this classes implies method to
		 * be called as soon as the this class is installed as the Policy
		 * implementation. The implies method cannot cause class loading to
		 * occur before there is the delegation to the non-JACC Policy provider
		 * or else an infinite recursion scenario arises where entrance into
		 * implies triggers class loading which recurses into implies. Here we
		 * load the JACC permission classes to ensure we get to the point of
		 * being able to delegate non-JACC permission to the delegate policy.
		 * This is the same type of statement performed at the start of implies
		 * which was causing the JACC permissions to be loaded. See [JBAS-1363].
		 */
		Permission permission = new RuntimePermission("test");
		boolean loadedPerms = !(permission instanceof EJBMethodPermission
				|| permission instanceof EJBRoleRefPermission
				|| permission instanceof WebResourcePermission
				|| permission instanceof WebRoleRefPermission || permission instanceof WebUserDataPermission);
		if (trace) {
			log.trace("Loaded JACC permissions: " + loadedPerms);
		}
		// Load PolicyContext as this also can trigger permission checks in
		// implies
		Class c = PolicyContext.class;
	}

	public Class[] getExternalPermissionTypes() {
		return externalPermissionTypes;
	}

	public void setExternalPermissionTypes(Class[] externalPermissionTypes) {
		if (externalPermissionTypes == null) {
			externalPermissionTypes = new Class[0];
		}
		this.externalPermissionTypes = externalPermissionTypes;
	}

	public PermissionCollection getPermissions(ProtectionDomain domain) {
		PermissionCollection pc = super.getPermissions(domain);
		PermissionCollection delegated = delegate.getPermissions(domain);
		for (Enumeration e = delegated.elements(); e.hasMoreElements();) {
			Permission p = (Permission) e.nextElement();
			pc.add(p);
		}
		return pc;
	}

	/**
	 * Return the permission collection associated with the cs. If there is no
	 * active JACC PolicyContext then the delegate value for
	 * getPermissions(CodeSource) is returned. Otherwise the JACC policy context
	 * permissions are returned.
	 * 
	 * @param cs -
	 *            the CodeSource
	 * @return the associated permission collection
	 */
	public PermissionCollection getPermissions(CodeSource cs) {
		PermissionCollection pc;
		String contextID = PolicyContext.getContextID();
		if (contextID == null) {
			pc = delegate.getPermissions(cs);
		} else {
			ContextPolicy policy = (ContextPolicy) instance.activePolicies
					.get(contextID);
			if (policy != null) {
				pc = policy.getPermissions();
			} else {
				pc = delegate.getPermissions(cs);
			}
		}
		return pc;
	}

	/**
	 * We dynamically manage the active policies on commit so refresh is a noop.
	 * Its not clear from the spec whether committed policies should not be
	 * visible until a refresh.
	 */
	public void refresh() {

	}

	/**
	 * returning 'this' for now (instead of the proxy) since I need access to
	 * more functionality in the security provider's start(). currently the
	 * policy is injected as an mbean attribute, hence we need an accessor like
	 * this one to get to it (since Policy is not an interface) todo: figure out
	 * how to protect this better and keep the functionality todo: or better:
	 * fix the DelegatingPolicy so that we can use it instead of this ;)
	 * 
	 * @return A proxy for our Policy interface
	 */
	public Policy getPolicyProxy() {
		return this;
	}

	// Policy configuration methods used by the PolicyConfiguration impl

	/**
	 * Access the current ContextPolicy instances
	 * 
	 * @return Map<String, ContextPolicy> of the contextID to policy mappings
	 */
	public String listContextPolicies() {
		StringBuffer tmp = new StringBuffer("<ActiveContextPolicies>");
		Iterator iter = activePolicies.keySet().iterator();
		while (iter.hasNext()) {
			String contextID = (String) iter.next();
			ContextPolicy cp = (ContextPolicy) activePolicies.get(contextID);
			tmp.append(cp);
			tmp.append('\n');
		}
		tmp.append("</ActiveContextPolicies>");

		tmp.append("<OpenContextPolicies>");
		iter = openPolicies.keySet().iterator();
		while (iter.hasNext()) {
			String contextID = (String) iter.next();
			ContextPolicy cp = (ContextPolicy) openPolicies.get(contextID);
			tmp.append(cp);
			tmp.append('\n');
		}
		tmp.append("</OpenContextPolicies>");

		return tmp.toString();
	}

	synchronized ContextPolicy getContextPolicy(String contextID)
			throws PolicyContextException {
		ContextPolicy policy = (ContextPolicy) openPolicies.get(contextID);
		if (policy == null) {
			throw new PolicyContextException(
					"No ContextPolicy exists for contextID=" + contextID);
		}
		return policy;
	}

	/**
	 * Create or update a ContextPolicy for contextID. This moves any active
	 * policy to the openPolicies map until its committed.
	 * 
	 * @param contextID
	 * @param remove
	 * @throws javax.security.jacc.PolicyContextException
	 * 
	 */
	synchronized void initPolicyConfiguration(String contextID, boolean remove)
			throws PolicyContextException {
		// Remove from the active policy map
		ContextPolicy policy = (ContextPolicy) activePolicies.remove(contextID);
		if (policy == null) {
			policy = (ContextPolicy) openPolicies.get(contextID);
		}
		if (policy == null) {
			policy = new ContextPolicy(contextID);
		}
		// Add to the open policy map
		openPolicies.put(contextID, policy);
		if (remove) {
			policy.clear();
		}
	}

	void addToExcludedPolicy(String contextID, Permission permission)
			throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.addToExcludedPolicy(permission);
	}

	void addToExcludedPolicy(String contextID, PermissionCollection permissions)
			throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.addToExcludedPolicy(permissions);
	}

	void addToRole(String contextID, String roleName, Permission permission)
			throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.addToRole(roleName, permission);
	}

	void addToRole(String contextID, String roleName,
			PermissionCollection permissions) throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.addToRole(roleName, permissions);
	}

	void addToUncheckedPolicy(String contextID, Permission permission)
			throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.addToUncheckedPolicy(permission);
	}

	void addToUncheckedPolicy(String contextID, PermissionCollection permissions)
			throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.addToUncheckedPolicy(permissions);
	}

	void linkConfiguration(String contextID, PolicyConfiguration link)
			throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		ContextPolicy linkPolicy = getContextPolicy(link.getContextID());
		policy.linkConfiguration(linkPolicy);
	}

	/**
	 * May need to make this synchronized to allow the move from the open to
	 * active policy map atomic. Right now the assumption is that a single
	 * thread is active for a given contextID.
	 * 
	 * @param contextID
	 * @throws javax.security.jacc.PolicyContextException
	 * 
	 */
	public void commit(String contextID) throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		openPolicies.remove(contextID);
		activePolicies.put(contextID, policy);
		policy.commit();
	}

	public void delete(String contextID) throws PolicyContextException {
		ContextPolicy policy = (ContextPolicy) activePolicies.remove(contextID);
		if (policy == null) {
			policy = (ContextPolicy) openPolicies.remove(contextID);
		}
		if (policy != null) {
			policy.delete();
		}
	}

	void removeExcludedPolicy(String contextID) throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.removeExcludedPolicy();
	}

	void removeRole(String contextID, String roleName)
			throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.removeRole(roleName);
	}

	void removeUncheckedPolicy(String contextID) throws PolicyContextException {
		ContextPolicy policy = getContextPolicy(contextID);
		policy.removeUncheckedPolicy();
	}

	public boolean implies(ProtectionDomain domain, Permission permission) {
		// Check the
		boolean isJaccPermission = permission instanceof EJBMethodPermission
				|| permission instanceof EJBRoleRefPermission
				|| permission instanceof WebResourcePermission
				|| permission instanceof WebRoleRefPermission
				|| permission instanceof WebUserDataPermission;
		boolean implied = false;
		// If there are external permission types check them
		if (!isJaccPermission && instance.externalPermissionTypes.length > 0) {
			Class pc = permission.getClass();
			for (int n = 0; n < instance.externalPermissionTypes.length; n++) {
				Class epc = instance.externalPermissionTypes[n];
				if (epc.isAssignableFrom(pc)) {
					isJaccPermission = true;
					break;
				}
			}
		}

		if (!isJaccPermission) {
			// Let the delegate policy handle the check
			implied = delegate.implies(domain, permission);
		} else {
			if (trace) {
				log.trace("implies, domain=" + domain + ", permission="
						+ permission);
				try {
					Subject caller = (Subject) PolicyContext
							.getContext("javax.security.auth.Subject.container");
					log.trace("implies javax.security.auth.Subject.container: "
							+ caller);
				} catch (Throwable e) {
					log.trace("Failed to access Subject context", e);
				}
			}
			String contextID = PolicyContext.getContextID();
			ContextPolicy contextPolicy = (ContextPolicy) instance.activePolicies
					.get(contextID);
			if (contextPolicy != null) {
				
				//do silent authorization check here
				if (silentAuthorization(permission)) {
					implied = true;					
				} else {
					implied = contextPolicy.implies(domain, permission);
				}
			} else if (trace) {
				log.trace("No PolicyContext found for contextID=" + contextID);
			}
		}
		if (trace) {
			log.trace("implied=" + implied);
		}

		return implied;
	}

	/**
	 * Check the external properties file and silently authorize
	 * the portal window, portlet and portlet instance objects.
	 * 
	 * Hard-coded to a simple properties file as a demonstration.
	 * A real implementation would check LDAP or Windows Active Directory,
	 * or a custom IDM solution.
	 * 
	 * @param permission
	 * @return
	 */
	private boolean silentAuthorization(Permission permission) {
		boolean silentAuth = false;
		// do spnego check
		String silentAuthRole = "Administrator";
		String permissionStr = stripPermissionString(permission.getName());		

		try {
			Properties props = new Properties();
			FileInputStream fsi = new FileInputStream(SILENT_AUTHORIZATION_STORE);
			props.load(fsi);
			StringTokenizer zer = new StringTokenizer(props
					.getProperty(silentAuthRole), ",");
			log.info("Permission Check for permission" + permissionStr);
			
			while (zer.hasMoreTokens()) {
				String token = zer.nextToken();
				if (permissionStr.equals(token))
				{
					silentAuth = true;
				}				
			}
		} catch (Exception e) {
			log.error("A custom security exception has ocurred", e);
		}

		return silentAuth;
	}
	
	/**
	 * Strips out the portal object naming conventions.
	 * 
	 * Whole point of this method is to enable custom authorization properties
	 * file to require a single entry for portal windows, portlets and instances.
	 * 
	 * It's rare that you'll grant access to a portal window, but not the portlet
	 * or portlet intances.
	 *  
	 * @param permissionStr
	 * @return
	 */
	private String stripPermissionString(String permissionStr)
	{
		permissionStr = permissionStr.replace("Window","");
		permissionStr = permissionStr.replace("Portlet","");
		permissionStr = permissionStr.replace("Instance","");	
		return permissionStr;
	}

	/**
	 * This proxy wrapper restricts the visible methods to only those from the
	 * Policy base class.
	 */
	public static class PolicyProxy extends Policy {
		private Policy delegate;

		PolicyProxy(Policy delegate) {
			this.delegate = delegate;
		}

		public void refresh() {
			delegate.refresh();
		}

		public PermissionCollection getPermissions(CodeSource codesource) {
			return delegate.getPermissions(codesource);
		}

		public boolean implies(ProtectionDomain domain, Permission permission) {
			return delegate.implies(domain, permission);
		}

		public PermissionCollection getPermissions(ProtectionDomain domain) {
			return delegate.getPermissions(domain);
		}
	}
}
